from artifician.processors.text_processors.stemlemtizer import StemLemProcessor
import pytest
class publisher:
    def __init__(self) -> None:
        self.value = ""

class TestStemlemitizer:
    def test_lemmatization(self):
        processor = StemLemProcessor(mode='lemmatization')
        pub = publisher()
        result = processor.process(pub, "The cats are running")
        assert result == "the cat be run"

    # Test 2: Test Stemming
    def test_stemming(self):
        processor = StemLemProcessor(mode='stemming')
        pub = publisher()
        result = processor.process(pub, "The cats are running")
        assert result == "the cat are run"

    # Test 3: Test None Input
    def test_none_input(self):
        processor = StemLemProcessor(mode='lemmatization')
        with pytest.raises(ValueError):
            processor.process(None, None)

    # Test 4: Test Empty String
    def test_empty_string(self):
        processor = StemLemProcessor(mode='lemmatization')
        pub = publisher()
        result = processor.process(pub, "")
        assert result == ""

    # Test 6: Test Invalid Mode
    def test_invalid_mode(self):
        with pytest.raises(ValueError):
            StemLemProcessor(mode='invalid')

    # Test 7: Test Single Word Lemmatization
    def test_single_word_lemmatization(self):
        processor = StemLemProcessor(mode='lemmatization')
        pub = publisher()
        result = processor.process(pub, "running")
        assert result == "run"

    # Test 8: Test Single Word Stemming
    def test_single_word_stemming(self):
        processor = StemLemProcessor(mode='stemming')
        pub = publisher()
        result = processor.process(pub, "running")
        assert result == "run"


from artifician.processors.processor import Processor
import spacy
from typing import List, Union
class TokenizationProcessor(Processor):
    """
    Tokenization Processor for splitting text into tokens.

    Attributes:
        method (str): Method to use for tokenization ('word' or 'sentence').
        nlp: spaCy language model for processing text.
    """


    def __init__(self, method: str = 'word', subscribe_to=None):
        """
        Initialize a TokenizationProcessor object.

        Args:
            method (str): Method to use for tokenization ('word' or 'sentence').
        """
        super().__init__()
        if method not in ['word', 'sentence']:
            raise ValueError("Invalid tokenization method. Choose 'word' or 'sentence'.")

        self.method = method
        self.nlp = spacy.load("en_core_web_sm")

        if subscribe_to is not None:
            for publisher in subscribe_to:
                self.subscribe(publisher)

    def process(self, publisher, text: Union[str, List[str], None]) -> Union[List[str], List[List[str]]]:
        """
        Process the text or list of texts and split it into tokens.

        Args:
            text (Union[str, List[str], None]): The text or list of texts to be tokenized.

        Returns:
            Union[List[str], List[List[str]]]: A list of tokens or list of lists of tokens.

        Raises:
            ValueError: If the input text is None or an empty list.
        """
        if text is None or (isinstance(text, list) and not text):
            raise ValueError("Input text cannot be None or an empty list for tokenization.")

        if isinstance(text, list):
            publisher.value = [self._tokenize(t) for t in text]
            return publisher.value
        else:
            return self._tokenize(text)

    def _tokenize(self, text: str) -> List[str]:
        """
        Internal method to tokenize a single text string.

        Args:
            text (str): The text to be tokenized.

        Returns:
            List[str]: A list of tokens.
        """
        if self.method == 'word':
            return [token.text for token in self.nlp(text)]
        elif self.method == 'sentence':
            return [sent.text.strip() for sent in self.nlp(text).sents]

        raise ValueError(f"Unknown tokenization method: {self.method}")
    
    def subscribe(self, publisher, pool_scheduler=None):
        """Defines logic for subscribing to an event in publisher

        Args:
            publisher (object): instance of the publisher
            pool_scheduler (rx.scheduler.ThreadPoolScheduler): scheduler instance for concurrency

        Returns:
            None
        """
        observable = publisher.observe(publisher.EVENT_PROCESSED)
        observable = observable.subscribe(lambda value: self.process(publisher, value),
                                          scheduler=pool_scheduler)


from sklearn.feature_extraction.text import TfidfVectorizer
from collections import Counter
from typing import List, Union, Optional
import logging
import spacy

logging.basicConfig(level=logging.INFO)
class KeywordExtractor:
    """
    Keyword Extractor class for extracting specific keywords from a text.

    Attributes:
        method (str): Method to use for keyword extraction ('manual', 'frequency', 'tfidf', etc.)
        keywords (List[str]): List of keywords to search within the text for 'manual' method.
    """

    def __init__(self, method: str = 'manual', keywords: List[str] = None):
        """
        Initialize a new KeywordExtractor object.

        Args:
            method (str): Method to use for keyword extraction.
            keywords (List[str]): List of keywords to search within the text for 'manual' method.

        Raises:
            ValueError: If the keywords list is empty for 'manual' method.
        """
        self.method = method
        if self.method == 'manual' and not keywords:
            raise ValueError("Keywords list cannot be empty for 'manual' method.")
        
        self.keywords = keywords if self.method == 'manual' else None
        self.nlp = spacy.load("en_core_web_sm")  # Load English tokenizer, POS tagger, NER, etc.

    def extract(self, text: Union[str, None], corpus: List[str] = None) -> List[str]:
        if text is None:
            raise ValueError("Input text cannot be None.")

        if self.method == 'manual':
            return self._manual_extract(text)
        elif self.method == 'frequency':
            return self._frequency_extract(text)
        elif self.method == 'tfidf':
            if corpus is None:
                raise ValueError("Corpus can't be None for 'tfidf' method.")
            return self._tfidf_extract(text, corpus)
        elif self.method == 'ngram':
            return self._ngram_extract(text)
        elif self.method == 'ner':
            return self._ner_extract(text)
        elif self.method == 'pos':
            return self._pos_extract(text)
        else:
            raise ValueError(f"Unknown method {self.method}")

    def _manual_extract(self, text: str) -> List[str]:
        """
        Manually extract keywords based on a pre-defined list.

        Args:
            text (str): The text to extract keywords from.

        Returns:
            List[str]: Extracted keywords that match the list.
        """
        return [word for word in self.keywords if word in text]

    def _frequency_extract(self, text: str, min_frequency: int = 2) -> List[str]:
        """
        Extract keywords based on frequency.

        Args:
            text (str): The text to extract keywords from.
            min_frequency (int): Minimum frequency for a word to be considered a keyword.

        Returns:
            List[str]: Extracted keywords based on frequency.
        """
        words = text.split()
        word_freq = Counter(words)
        return [word for word, freq in word_freq.items() if freq >= min_frequency]

    def _tfidf_extract(self, text: str, corpus: List[str]) -> List[str]:
        """
        Extract keywords using TF-IDF.

        Args:
            text (str): The text to extract keywords from.
            corpus (List[str]): The collection of documents to fit the TF-IDF model.

        Returns:
            List[str]: Extracted keywords based on TF-IDF scores.
        """
        vectorizer = TfidfVectorizer()
        X = vectorizer.fit_transform(corpus)
        feature_names = vectorizer.get_feature_names_out()
        tfidf_vector = vectorizer.transform([text])
        sorted_items = self._sort_coo(tfidf_vector.tocoo())
        keywords = self._extract_topn_from_vector(feature_names, sorted_items, 10)
        return keywords

    @staticmethod
    def _sort_coo(coo_matrix):
        tuples = zip(coo_matrix.col, coo_matrix.data)
        return sorted(tuples, key=lambda x: (x[1], x[0]), reverse=True)

    @staticmethod
    def _extract_topn_from_vector(feature_names, sorted_items, topn=10):
        """return n-gram list with highest tf-idf."""

        # use only topn items from vector
        sorted_items = sorted_items[:topn]

        score_vals = []
        feature_vals = []

        for idx, score in sorted_items:
            fname = feature_names[idx]
            
            # keep track of feature name and its corresponding score
            score_vals.append(round(score, 3))
            feature_vals.append(feature_names[idx])

        results = {}
        for idx in range(len(feature_vals)):
            results[feature_vals[idx]] = score_vals[idx]
        
        return results

    def _ngram_extract(self, text: str, n=2) -> List[str]:
        """
        Extract n-grams from the text.

        Args:
            text (str): The text to extract n-grams from.
            n (int): The 'n' in n-grams, representing the number of grouped words.

        Returns:
            List[str]: Extracted n-grams.
        """
        words = text.split()
        ngrams = zip(*[words[i:] for i in range(n)])
        return [" ".join(ngram) for ngram in ngrams]
    
    def _ner_extract(self, text: str) -> List[str]:
        """
        Perform Named Entity Recognition (NER) to extract keywords.

        This method uses NER to identify and extract named entities such as names of people,
        organizations, locations, etc., from the given text.

        Args:
            text (str): The text from which to extract named entities.

        Returns:
            List[str]: A list of extracted named entities.

        Raises:
            Exception: If any error occurs during the extraction process.
        """
        try:
            doc = self.nlp(text)
            if not doc.ents:
                logging.warning("No named entities found in the text.")
                return []
            extracted_entities = [ent.text for ent in doc.ents]
            logging.info(f"NER extraction found {len(extracted_entities)} keywords.")
            return extracted_entities
        except Exception as e:
            logging.error(f"An error occurred during NER extraction: {e}")
            return []

    def _pos_extract(self, text: str, pos_tags: Optional[List[str]] = None) -> List[str]:
        """
        Extract keywords based on Part-of-Speech (POS) tagging.

        This method uses POS tagging to identify and extract specific types of words,
        such as nouns, proper nouns, and verbs, from the given text.

        Args:
            text (str): The text from which to extract words based on their POS tags.
            pos_tags (Optional[List[str]]): A list of POS tags to consider for extraction. 
                                            Defaults to ["NOUN", "PROPN", "VERB"].

        Returns:
            List[str]: A list of extracted words based on their POS tags.

        Raises:
            Exception: If any error occurs during the extraction process.
        """
        if pos_tags is None:
            pos_tags = ["NOUN", "PROPN", "VERB"]

        try:
            doc = self.nlp(text)
            extracted_words = [token.text for token in doc if token.pos_ in pos_tags]
            if not extracted_words:
                logging.warning(f"No words found with POS tags: {pos_tags}")
                return []
            logging.info(f"POS extraction found {len(extracted_words)} keywords.")
            return extracted_words
        except Exception as e:
            logging.error(f"An error occurred during POS extraction: {e}")
            return []


import pytest
from artifician.processors.text_processors.tokenizer import TokenizationProcessor

# Test 1: Test Word Tokenization

class publisher:
    def __init__(self) -> None:
        self.value = ""
class TestTokenizationProcessor:
    def test_word_tokenization(self,):
        tokenizer = TokenizationProcessor(method='word')
        pub = publisher()
        result = tokenizer.process(pub, "Hello world!")
        assert result == ["Hello", "world", "!"]

    # Test 2: Test Sentence Tokenization
    def test_sentence_tokenization(self):
        tokenizer = TokenizationProcessor(method='sentence')
        pub = publisher()
        result = tokenizer.process(pub,"Hello world! This is a test.")
        assert result == ["Hello world!", "This is a test."]

    # Test 3: Test None Input
    def test_none_input(self):
        tokenizer = TokenizationProcessor(method='word')
        with pytest.raises(ValueError):
            tokenizer.process(None, None)

    # Test 4: Test Empty String
    def test_empty_string(self):
        tokenizer = TokenizationProcessor(method='word')
        pub = publisher()
        result = tokenizer.process(pub,"")
        assert result == []

    # Test 5: Test List of Strings Input for Word Tokenization
    def test_list_of_strings_word_tokenization(self):
        tokenizer = TokenizationProcessor(method='word')
        pub = publisher()
        result = tokenizer.process(pub, ["Hello world!", "This is a test."])
        assert result == [["Hello", "world", "!"], ["This", "is", "a", "test", "."]]

    # Test 6: Test Invalid Tokenization Method
    def test_invalid_tokenization_method(self):
        with pytest.raises(ValueError):
            TokenizationProcessor(method='invalid')

    # Test 7: Test Punctuation Handling in Word Tokenization
    def test_punctuation_handling_word_tokenization(self):
        tokenizer = TokenizationProcessor(method='word')
        pub = publisher()
        result = tokenizer.process(pub,"Hello, world!")
        assert result == ["Hello", ",", "world", "!"]

    # Test 8: Test Multiple Sentences in Sentence Tokenization
    def test_multiple_sentences_sentence_tokenization(self):
        tokenizer = TokenizationProcessor(method='sentence')
        pub = publisher()
        result = tokenizer.process(pub,"Hello world! This is a test. Another sentence.")
        assert result == ["Hello world!", "This is a test.", "Another sentence."]


from artifician import *


import pytest

class TestProcessorChainManager:

    # can add processors to the chain using 'then' method
    def test_add_processors_to_chain(self):
        normalizer = Normalizer(PropertiesNormalizer(), delimiter = {'delimiter': [" "]})
        mapper = Mapper(feature_map=FeatureMap({"": 0}))
        manager = normalizer.then(mapper)

        assert len(manager.processors) == 2
        assert manager.processors[0] == normalizer
        assert manager.processors[1] == mapper

    # can process data sequentially through the chain of processors using 'process' method
    def test_process_data_sequentially(self):
        normalizer = Normalizer(PropertiesNormalizer(), delimiter = {'delimiter': [" "]})
        mapper = Mapper(feature_map=FeatureMap({"": 0}))
        manager = normalizer.then(mapper)
        data = "data"
        publisher = Dataset()
        result = manager.process(publisher, data)
        
        assert result == [1]


    # raises a TypeError when adding a processor without a 'process' method
    def test_add_processor_without_process_method(self):
        class ProcessorWithoutProcessMethod():
            def __init__(self):
                super().__init__()
        processor = ProcessorWithoutProcessMethod()
        manager = ProcessorChainManager()

        with pytest.raises(TypeError):
            manager.then(processor)

    # raises a RuntimeError when processing an empty processor chain
    def test_process_empty_chain(self):
        manager = ProcessorChainManager()
        publisher = Dataset()
        data = "data"

        with pytest.raises(RuntimeError):
            manager.process(publisher, data)



import pytest
from artifician.processors.text_processors.stop_word_remover import StopWordsRemoverProcessor


class publisher:
    def __init__(self) -> None:
        self.value = ""
# Test 1: Test Stop Word Removal with Single String
def test_stop_word_removal_single_string():
    processor = StopWordsRemoverProcessor()
    pub = publisher()
    result = processor.process(pub, "This is a test")
    assert result == "test"

# Test 2: Test None Input
def test_none_input():
    processor = StopWordsRemoverProcessor()
    pub = publisher()
    with pytest.raises(ValueError):
        processor.process(pub, None)

# Test 3: Test Empty String
def test_empty_string():
    processor = StopWordsRemoverProcessor()
    pub = publisher()
    result = processor.process(pub, "")
    assert result == ""

# Test 5: Test Custom Stop Words
def test_custom_stop_words():
    custom_stop_words = ["sample"]
    pub = publisher()
    processor = StopWordsRemoverProcessor(custom_stop_words=custom_stop_words)
    result = processor.process(pub, "This is a sample text")
    assert result == "text"

# Test 6: Test Single Word
def test_single_word():
    processor = StopWordsRemoverProcessor()
    pub = publisher()
    result = processor.process(pub, "sample")
    assert result == "sample"


from bs4 import BeautifulSoup
from artifician import *
from artifician.extractors.html_extractors import get_node_attribute

def test_processor_chain_with_init_method():
    html_elements = BeautifulSoup("""<div class="container"> <p class="Hello World">Hello</p>
                    <span class="Hello value">John</span> </div> """, "html.parser").find_all()

    dataset = Dataset()
    feature = FeatureDefinition(get_node_attribute, [dataset],  "class")

    text_normalizer = Normalizer(PropertiesNormalizer(), delimiter = {'delimiter': [" "]})
    mapper = Mapper(feature_map=FeatureMap({"": 0}))

    chain = ProcessorChainManager([text_normalizer, mapper])

    chain.subscribe(feature)
    expected_ouput = [[1], [2,3], [2,4]]
    datastore = dataset.add_samples(list(html_elements))
    assert expected_ouput == list(datastore[1].values)

def test_processor_chain_with_then_method():
    html_elements = BeautifulSoup("""<div class="container"> <p class="Hello World">Hello</p>
                    <span class="Hello value">John</span> </div> """, "html.parser").find_all()

    dataset = Dataset()
    feature = FeatureDefinition(get_node_attribute, [dataset],  "class")

    text_normalizer = Normalizer(PropertiesNormalizer(), delimiter = {'delimiter': [" "]})
    mapper = Mapper(feature_map=FeatureMap({"": 0}))

    chain = text_normalizer.then(mapper)

    chain.subscribe(feature)
    expected_ouput = [[1], [2,3], [2,4]]
    datastore = dataset.add_samples(list(html_elements))
    assert expected_ouput == list(datastore[1].values)


import spacy
from typing import List, Union
from artifician.processors.processor import Processor

class StopWordsRemoverProcessor(Processor):
    """
    Processor for removing stop words from text data.

    Attributes:
        stop_words (set): A set of stop words to be removed.
    """

    def __init__(self, custom_stop_words: List[str] = None, subscribe_to=None):
        """
        Initialize a StopWordsRemoverProcessor object.

        Args:
            custom_stop_words (List[str]): Optional list of custom stop words.
            subscribe_to (list): Optional list of publishers to subscribe to.
        """
        super().__init__()
        self.nlp = spacy.load("en_core_web_sm")
        self.stop_words = set(self.nlp.Defaults.stop_words)
        
        if custom_stop_words is not None:
            self.stop_words.update(custom_stop_words)

        if subscribe_to is not None:
            for publisher in subscribe_to:
                self.subscribe(publisher)

    def process(self, publisher, text: Union[str, List[str]]) -> Union[str, List[str]]:
        """
        Process the text or list of texts to remove stop words.

        Args:
            publisher: The publisher associated with the processor.
            text (Union[str, List[str]]): The text or list of texts to be processed.

        Returns:
            Union[str, List[str]]: Text after stop words removal.

        Raises:
            ValueError: If the input text is None or an empty list.
        """
        if text is None or (isinstance(text, list) and not text):
            raise ValueError("Input text cannot be None or an empty list.")

        if isinstance(text, list):
            processed_output = [word for word in text if word.lower() not in self.stop_words]
        else:
            processed_output = ' '.join(word for word in text.split() if word.lower() not in self.stop_words)

        publisher.value = processed_output
        return processed_output

    def subscribe(self, publisher, pool_scheduler=None):
        """
        Subscribe to a publisher for event-driven processing.

        Args:
            publisher (object): The publisher to subscribe to.
            pool_scheduler (optional): Scheduler instance for concurrency.

        Returns:
            None
        """
        observable = publisher.observe(publisher.EVENT_PROCESSED)
        observable.subscribe(lambda value: self.process(publisher, value),
                             scheduler=pool_scheduler)



"""
   Copyright 2023 Plato Solutions, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import pandas as pd
import logging
from typing import Any, List
from rx.subject import Subject

class Dataset:
    """
    Dataset contains all the functionality for preparing Artifician data.
    It observes events and stores all processed data in a Pandas DataFrame.

    Attributes:
        cached (dict): Cached observables for different events.
        datastore (pd.DataFrame): DataFrame to store all samples.
        PREPARE_DATASET (Callable): Event to prepare the dataset.
        POST_PROCESS (Callable): Event for post-processing actions on the dataset.
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.features_sample = None
        self.cached = {}
        self.datastore = pd.DataFrame()
        self.PREPARE_DATASET = self.add_samples
        self.POST_PROCESS = self.post_process

    def add_samples(self, samples: Any) -> pd.DataFrame:
        """
        Adds samples to the datastore.

        Args:
            samples (Any): Artifician data to be added.

        Returns:
            pd.DataFrame: The updated dataset.

        Raises:
            TypeError: If the input data is not a list.
        """
        if not isinstance(samples, list):
            self.logger.error(f"Input data should be list, not {type(samples)}")
            raise TypeError(f"Input data should be list, not {type(samples)}")

        sample_data = []

        try:
            for sample in samples:
                self.features_sample = [sample]
                if self.add_samples in self.cached:
                    self.cached[self.add_samples].on_next(sample)
                sample_data.append(self.features_sample)
        except KeyError as e:
            self.logger.warning("No module has subscribed to dataset: " + str(e))

        self.datastore = pd.DataFrame(sample_data)
        return self.datastore

    def observe(self, event):
        """
        Builds and returns an observable for a given event.

        Args:
            event (Callable): Function to create an observable from.

        Returns:
            rx.subject.Subject: Observable for the given event.
        """
        if event not in self.cached:
            self.cached[event] = Subject()

        return self.cached[event]

    def post_process(self):
        """
        This event should be called after Artifician data is prepared.
        Listeners to the post_process event can perform collective actions on the dataset.
        """
        if self.post_process in self.cached:
            self.cached[self.post_process].on_next(self.datastore)




"""
   Copyright 2023 Plato Solutions, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import logging
from abc import ABC, abstractmethod
from artifician.processors.processor_chain_manager import ProcessorChainManager

class Processor(ABC):
    """
    Interface for processors in the Artifician library, updated for processor chaining.

    This abstract class defines the interface for processors, including methods for processing data
    and subscribing to publishers, along with the ability to chain processors.
    """

    def __init__(self):
        self.next_processor = None
        self.chain_manager = None
        self.logger = logging.getLogger(__name__)

    @abstractmethod
    def process(self, publisher, *data):
        """
        Process the data and update the publisher with the processed values.

        Args:
            publisher: The publisher to which the processed data will be updated.
            data: The data to be processed.
        """
        pass

    @abstractmethod
    def subscribe(self, publisher, pool_scheduler=None):
        """
        Subscribe the processor to a publisher (e.g., FeatureDefinition).

        Args:
            publisher: The publisher to subscribe to.
            pool_scheduler (optional): The scheduler to be used for subscription.
        """
        pass

    def then(self, next_processor):
        """
        Link this processor to the next one in the chain.

        Args:
            next_processor: The next processor to add to the chain.

        Returns:
            ProcessorChainManager: The chain manager managing this processor chain.

        Raises:
            TypeError: If the next_processor is not a valid processor instance.
        """
        try:
            # Assuming all processors have a 'process' method
            if not hasattr(next_processor, 'process'):
                raise TypeError("The next_processor must be a valid processor instance")

            if not self.chain_manager:
                self.chain_manager = ProcessorChainManager([self])

            self.next_processor = next_processor
            self.chain_manager.then(next_processor)
            return self.chain_manager

        except Exception as e:
            self.logger.exception(f"Error in chaining processor: {e}")
            raise



import re
from artifician.processors.processor import Processor
from typing import List, Union

class TextCleaningProcessor(Processor):
    """
    Processor for cleaning and preprocessing text data.

    Configurable attributes for various cleaning operations.
    """

    def __init__(self, lowercase=True, remove_punctuation=True, remove_numbers=True, 
                 strip_whitespace=True, remove_html_tags=True, remove_urls=True, 
                 subscribe_to=None):
        """
        Initialize a TextCleaningProcessor object.

        Args:
            lowercase (bool): Flag to convert text to lowercase.
            remove_punctuation (bool): Flag to remove punctuation.
            remove_numbers (bool): Flag to remove numbers.
            strip_whitespace (bool): Flag to strip extra whitespaces.
            remove_html_tags (bool): Flag to remove HTML tags.
            remove_urls (bool): Flag to remove URLs.
            custom_stop_words (List[str]): Optional list of custom stop words.
            subscribe_to (list): Optional list of publishers to subscribe to.
        """
        super().__init__()
        self.lowercase = lowercase
        self.remove_punctuation = remove_punctuation
        self.remove_numbers = remove_numbers
        self.strip_whitespace = strip_whitespace
        self.remove_html_tags = remove_html_tags
        self.remove_urls = remove_urls

        if subscribe_to is not None:
            for publisher in subscribe_to:
                self.subscribe(publisher)

    def process(self, publisher, text: Union[str, List[str]]) -> Union[str, List[str]]:
        """
        Process the text or list of texts to clean and preprocess.

        Args:
            publisher: The publisher associated with the processor.
            text (Union[str, List[str]]): The text or list of texts to be processed.

        Returns:
            Union[str, List[str]]: Cleaned and preprocessed text.
        """
        if text is None or (isinstance(text, list) and not text):
            raise ValueError("Input text cannot be None or an empty list.")

        processed_output = [self._clean_text(t) for t in text] if isinstance(text, list) else self._clean_text(text)
        publisher.value = processed_output
        return processed_output

    def _clean_text(self, text: str) -> str:
        """
        Internal method to apply configured cleaning operations to a single text string.

        Args:
            text (str): The text to be cleaned.

        Returns:
            str: Cleaned text.
        """
        if self.lowercase:
            text = text.lower()
        if self.remove_urls:
            text = re.sub(r'https?://\S+|www\.\S+', '', text)
        if self.remove_html_tags:
            text = re.sub(r'<.*?>', '', text)
        if self.remove_punctuation:
            text = re.sub(r'[^\w\s]', '', text)
        if self.remove_numbers:
            text = re.sub(r'\d+', '', text)
        if self.strip_whitespace:
            text = text.strip()
            text = re.sub(r'\s+', ' ', text)
        return text


    def subscribe(self, publisher, pool_scheduler=None):
        """
        Subscribe to a publisher for event-driven processing.

        Args:
            publisher (object): The publisher to subscribe to.
            pool_scheduler (optional): Scheduler instance for concurrency.

        Returns:
            None
        """
        observable = publisher.observe(publisher.EVENT_PROCESSED)
        observable.subscribe(lambda value: self.process(publisher, value),
                             scheduler=pool_scheduler)



from typing import List, Union
import spacy
from nltk.stem import PorterStemmer
from artifician.processors.processor import Processor



class StemLemProcessor(Processor):
    """
    Processor for applying stemming and lemmatization to text data.

    Attributes:
        mode (str): Mode of operation ('stemming' or 'lemmatization').
        nlp: spaCy language model for lemmatization.
        stemmer: NLTK stemmer for stemming.
    """

    def __init__(self, mode: str = 'lemmatization', subscribe_to=None):
        """
        Initialize a StemLemProcessor object.

        Args:
            mode (str): Operation mode ('stemming' or 'lemmatization').
            subscribe_to (list): Optional list of publishers to subscribe to.
        """
        super().__init__()
        if mode not in ['stemming', 'lemmatization']:
            raise ValueError("Invalid mode. Choose 'stemming' or 'lemmatization'.")

        self.mode = mode
        self.nlp = spacy.load("en_core_web_sm") if mode == 'lemmatization' else None
        self.stemmer = PorterStemmer() if mode == 'stemming' else None

        if subscribe_to is not None:
            for publisher in subscribe_to:
                self.subscribe(publisher)

    def process(self, publisher, text: Union[str, List[str]]) -> Union[str, List[str]]:
        """
        Process the text or list of tokens for stemming or lemmatization.

        Args:
            publisher: The publisher associated with the processor.
            text (Union[str, List[str]]): The text or list of tokens to be processed.

        Returns:
            Union[str, List[str]]: Processed text or list of processed tokens.
        """
        if text is None or (isinstance(text, list) and not text):
            raise ValueError("Input text cannot be None or an empty list.")

        if isinstance(text, list):
            processed_tokens = [self._process_token(token) for token in text]
            publisher.value = processed_tokens
            return processed_tokens
        else:
            processed_text = ' '.join([self._process_token(token) for token in text.split()])
            publisher.value = processed_text
            return processed_text

    def _process_token(self, token: str) -> str:
        """
        Internal method to apply stemming or lemmatization to a single token.

        Args:
            token (str): The token to be processed.

        Returns:
            str: Processed token.
        """
        if self.mode == 'lemmatization':
            return self.nlp(token)[0].lemma_
        elif self.mode == 'stemming':
            return self.stemmer.stem(token)

    def subscribe(self, publisher, pool_scheduler=None):
        """
        Subscribe to a publisher for event-driven processing.

        Args:
            publisher (object): The publisher to subscribe to.
            pool_scheduler (optional): Scheduler instance for concurrency.

        Returns:
            None
        """
        observable = publisher.observe(publisher.EVENT_PROCESSED)
        observable.subscribe(lambda value: self.process(publisher, value),
                             scheduler=pool_scheduler)



"""
   Copyright 2023 Plato Solutions, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import logging
from typing import Tuple, Any, Callable, List
from rx.subject import Subject

class FeatureDefinition:
    """
    Contains all the functionality for preparing a single feature.

    Attributes:
        value (Any): The value of the feature.
        cached (dict): Cached observables for different events.
        extractor (Callable): Function to extract feature value from the artifician.
        EVENT_PROCESSED (Callable): Event that processes the feature.
        MAP_VALUES (Callable): Event that maps values of the feature.
        extractor_parameters (Tuple): Parameters for the extractor function.
    """

    def __init__(self, extractor: Callable = lambda sample: sample, subscribe_to: List = None, *extractor_parameters):
        """
        Initializes a FeatureDefinition instance.

        Args:
            extractor (Callable, optional): Function to extract feature value.
            subscribe_to (List): List of publishers to subscribe to.
            extractor_parameters: Additional parameters for the extractor.

        Raises:
            ValueError: If no publishers are provided to subscribe to.
        """
        self.logger = logging.getLogger(__name__)

        if not subscribe_to:
            self.logger.error("FeatureDefinition must be subscribed to at least one publisher.")
            raise ValueError("FeatureDefinition must be subscribed to at least one publisher.")

        self.value = None
        self.cached = {}
        self.extractor = extractor
        self.extractor_parameters = extractor_parameters
        self.EVENT_PROCESSED = self.process
        self.MAP_VALUES = self.map

        for publisher in subscribe_to:
            self.subscribe(publisher)

    def process(self, publisher, sample: Any) -> None:
        """
        Processes the sample to build the feature value.

        Args:
            sample (Any): The sample data.
            publisher: The instance of the publisher.
        """
        try:
            self.value = self.extractor(sample, *self.extractor_parameters)

            if self.process in self.cached:
                self.cached[self.process].on_next(self.value)

            self.map(self.value)
            publisher.features_sample.append(self.value)
        except Exception as e:
            self.logger.exception(f"Error in processing feature: {e}")
            raise

    def map(self, feature_value: Any) -> None:
        """
        Maps the feature value into an int or list of ints.

        Args:
            feature_value (Any): The feature value to be mapped.
        """
        try:
            if self.map in self.cached:
                self.cached[self.map].on_next(feature_value)
        except Exception as e:
            self.logger.exception(f"Error in mapping feature value: {e}")
            raise

    def observe(self, event: Callable) -> Subject:
        """
        Builds and returns an observable for a given event.

        Args:
            event (Callable): The function to create an observable from.

        Returns:
            Observable: An observable for the given event.
        """
        if event not in self.cached:
            self.cached[event] = Subject()

        return self.cached[event]

    def subscribe(self, publisher, pool_scheduler=None) -> None:
        """
        Defines logic for subscribing to an event in a publisher.

        Args:
            publisher: The publisher instance.
            pool_scheduler (optional): The scheduler instance for concurrency.
        """
        try:
            observable = publisher.observe(publisher.PREPARE_DATASET)
            observable.subscribe(lambda feature_raw: self.process(publisher, feature_raw),
                                 scheduler=pool_scheduler)
        except AttributeError as e:
            self.logger.error(f"Failed to subscribe to publisher: {e}")
            raise



from artifician.processors.text_processors.text_cleaner import TextCleaningProcessor

class publisher:
    def __init__(self) -> None:
        self.value = ""
# Test 1: Test Lowercasing
def test_lowercasing():
    pub = publisher()
    processor = TextCleaningProcessor(lowercase=True)
    assert processor.process(pub, "Sample TEXT") == "sample text"

# Test 2: Test Removing Punctuation
def test_remove_punctuation():
    pub = publisher()
    processor = TextCleaningProcessor(remove_punctuation=True)
    assert processor.process(pub, "Hello, world!") == "hello world"

# Test 3: Test Removing Numbers
def test_remove_numbers():
    pub = publisher()
    processor = TextCleaningProcessor(remove_numbers=True)
    assert processor.process(pub, "There are 4 apples") == "there are apples"

# Test 4: Test Stripping Whitespace
def test_strip_whitespace():
    pub = publisher()
    processor = TextCleaningProcessor(strip_whitespace=True)
    assert processor.process(pub, "  Text with spaces  ") == "text with spaces"

# Test 5: Test Removing HTML Tags
def test_remove_html_tags():
    pub = publisher()
    processor = TextCleaningProcessor(remove_html_tags=True)
    assert processor.process(pub, "<p>Some text</p>") == "some text"

# Test 6: Test Removing URLs
def test_remove_urls():
    pub = publisher()
    processor = TextCleaningProcessor(remove_urls=True)
    assert processor.process(pub, "Visit https://example.com") == "visit"

# Test 7: Test List of Strings
def test_list_of_strings():
    pub = publisher()
    processor = TextCleaningProcessor(lowercase=True, remove_punctuation=True)
    assert processor.process(pub, ["Hello, World!", "Sample TEXT"]) == ["hello world", "sample text"]



from artifician.processors.text_processors.stemlemtizer import *
from artifician.processors.text_processors.stop_word_remover import *
from artifician.processors.text_processors.tokenizer import *
from artifician.processors.text_processors.text_cleaner import *


"""
   Copyright 2023 Plato Solutions, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
class ProcessorChainManager:
    """Manages a chain of processors.

    This class handles the sequential execution of a chain of processors and
    can subscribe to a publisher to trigger the processing.

    Attributes:
        processors (list): A list of processors in the chain.
    """
    def __init__(self, processors=None) -> None:
        """Initializes the ProcessorChainManager with an optional list of processors.

        Args:
            processors (list, optional): An initial list of processors to be managed.
        """
        self.processors = processors or []

    def then(self, next_processor) -> 'ProcessorChainManager':
        """Adds a processor to the end of the chain.

        Args:
            processor (Processor): The processor to add to the chain.

        Returns:
            processor_chain_manager (ProcessorChainManager): The ProcessorChainManager instance.
        """
        if not hasattr(next_processor, 'process'):
            raise TypeError("next_processor must have a 'process' method")
        
        self.processors.append(next_processor)
        
        return self


    def process(self, publisher, data: any) -> any:
        """Processes data sequentially through the chain of processors.

        Args:
            data: The data to be processed by the chain.

        Returns:
            The final processed data after passing through all processors.
        """
        if not self.processors:
            raise RuntimeError("Processor chain is empty. Add processors before processing.")
        processed_data = data
        
        for processor in self.processors:
            try:
                processed_data = processor.process(publisher, processed_data)
            except Exception as e:
                raise RuntimeError(f"Error processing data in {processor}: {e}")
        
        publisher.value = processed_data

        return processed_data

    def subscribe(self, publisher, pool_scheduler=None) -> None:
        """Subscribes the processor chain to a feature definition.

        The feature definition will trigger the processing of the chain.

        Args:
            feature_definition (publisher): The feature definition to subscribe to.
        """
        observable = publisher.observe(publisher.EVENT_PROCESSED)
        observable.subscribe(lambda feature_raw: self.process(publisher, feature_raw),
                             scheduler=pool_scheduler)